% This function runs a conventional SE using Weighted Least Square Method

function [ result, h, r, r_N, J, x] = wlsfunction(zdata, ybus, bsh, g, b, x)

fbus = zdata(:,1); % from bus
tbus = zdata(:,2); % to bus
nbus = max([fbus;tbus]); % number of buses
type = zdata(:,3); % type of measurement
z = zdata(:,4); % measurement values

Ri = diag(zdata(:,5)); % measurement variance matrix
W = Ri\eye(size(Ri)); 
B = imag(ybus);
G = real(ybus);

%Initiating state variables
if exist('x', 'var') == 0
    V = ones(nbus,1); % Initialize the bus voltages..
    O = zeros(nbus,1);
    x = [O(2:end); V];   % State Vector, not including Theta_1
else
    O = [0; x(1:(nbus-1))];
    V = x(nbus:end);
end

%Indexes in zdata of the different type of measurements:
vi = find(type == 1); % Index of voltage magnitude measurements
ppi = find(type == 2); % Index of active power injection measurements
qi = find(type == 3); % Index of reactive power injection measurements
pf = find(type == 4); % Index of active power flow measurements
qf = find(type == 5); % Index of reactive power flow measurements
%vir = find(type == 6); % Index of real voltage measurements
%vij = find(type == 7); % Index of imaginary voltage measurements
ifr = find(type == 8); % Index of real current measurements
ifj = find(type == 9); % Index of imaginary current measurements
adi = find(type == 10); %Index of delta differences
vdi = find(type == 11); %Index of voltage differences

%Number of measurements of the different types:
nvi = length(vi); % Number of voltage measurements
npi = length(ppi); % Number of real power injection measurements
nqi = length(qi); % Number of reactive power injection measurements
npf = length(pf); % Number of real power flow measurements
nqf = length(qf); % Number of reactive power flow measurements
%nvir = length(vir); % Number of real voltage measurements
%nvij = length(vij); % Number of imaginary voltage measurements
nifr = length(ifr); % Number of real current measurements
nifj = length(ifj); % Number of imaginary current measurements
nadi = length(adi); %
nvdi = length(vdi); %

iter = 1;
tol = 5;
while(tol > 1e-5 && iter < 15)
    %% Building h vector
    h1 = V(fbus(vi),1);
    h2 = zeros(npi,1);
    h3 = zeros(nqi,1);
    h4 = zeros(npf,1);
    h5 = zeros(nqf,1);
    %h6 = V(fbus(vir)).*cos(O(fbus(vir)));
    %h7 = V(fbus(vij)).*sin(O(fbus(vij)));
    h8 = zeros(nifr, 1);
    h9 = zeros(nifj, 1);
    
    for i = 1:npi
        m = fbus(ppi(i));
        for k = 1:nbus
            h2(i) = h2(i) + V(m)*V(k)*(G(m,k)*cos(O(m)-O(k)) + B(m,k)*sin(O(m)-O(k)));
        end
    end
    
    for i = 1:nqi
        m = fbus(qi(i));
        for k = 1:nbus
            h3(i) = h3(i) + V(m)*V(k)*(G(m,k)*sin(O(m)-O(k)) - B(m,k)*cos(O(m)-O(k)));
        end
    end
    
    for i = 1:npf
        m = fbus(pf(i));
        n = tbus(pf(i));
        h4(i) = V(m)^2*g(m,n) - V(m)*V(n)*(g(m,n)*cos(O(m)-O(n)) + b(m,n)*sin(O(m)-O(n)));
    end
    
    for i = 1:nqf
        m = fbus(qf(i));
        n = tbus(qf(i));
        h5(i) = -V(m)^2*(b(m,n)+bsh(m,n)) - V(m)*V(n)*(g(m,n)*sin(O(m)-O(n)) - b(m,n)*cos(O(m)-O(n)));
    end
    
    for i = 1:nifr
        m = fbus(ifr(i));
        n = tbus(ifr(i));
        h8(i) = ( V(m)*cos(O(m))-V(n)*cos(O(n)) )*g(m,n)  - ( V(m)*sin(O(m)) - V(n)*sin(O(n)) )*b(m,n) - bsh(1,2)*V(m)*sin(O(m));
    end
    
    for i = 1:nifj
        m = fbus(ifj(i));
        n = tbus(ifj(i));
        h9(i) = ( ( V(m)*cos(O(m))-V(n)*cos(O(n)) )*b(m,n)  + ( V(m)*sin(O(m)) - V(n)*sin(O(n)) )*g(m,n) + bsh(m,n)*V(m)*cos(O(m)));
    end
    
    h = [h1; h2; h3; h4; h5; h8; h9];
    
    r = z - h; % residual 
   
    %% Building H matrix
    
    % H11 - Derivative of V with resepect to phase angles
    H11 = zeros(nvi,nbus-1);

    % H12 - Derivative of V with respect to V
    H12 = zeros(nvi,nbus);
    for k = 1:nvi
        H12(k,fbus(vi(k))) = 1;
    end

    % H21 - Derivative of Real Power Injections with resepect to phase angles
    H21 = zeros(npi,nbus-1);
    for i = 1:npi
        m = fbus(ppi(i));
        for k = 1:(nbus-1)
            if k+1 == m
                for n = 1:nbus
                    H21(i,k) = H21(i,k) + V(m)* V(n)*(-G(m,n)*sin(O(m)-O(n)) + B(m,n)*cos(O(m)-O(n)));
                end
                H21(i,k) = H21(i,k) - V(m)^2*B(m,m);
            else
                H21(i,k) = V(m)* V(k+1)*(G(m,k+1)*sin(O(m)-O(k+1)) - B(m,k+1)*cos(O(m)-O(k+1)));
            end
        end
    end
    
    % H22 - Derivative of Real Power Injections with V
    H22 = zeros(npi,nbus);
    for i = 1:npi
        m = fbus(ppi(i));
        for k = 1:(nbus)
            if k == m
                for n = 1:nbus
                    H22(i,k) = H22(i,k) + V(n)*(G(m,n)*cos(O(m)-O(n)) + B(m,n)*sin(O(m)-O(n)));
                end
                H22(i,k) = H22(i,k) + V(m)*G(m,m);
            else
                H22(i,k) = V(m)*(G(m,k)*cos(O(m)-O(k)) + B(m,k)*sin(O(m)-O(k)));
            end
        end
    end
    
    % H31 - Derivative of Reactive Power Injections with resepect to phase angles
    H31 = zeros(nqi,nbus-1);
    for i = 1:nqi
        m = fbus(qi(i));
        for k = 1:(nbus-1)
            if k+1 == m
                for n = 1:nbus
                    H31(i,k) = H31(i,k) + V(m)* V(n)*(G(m,n)*cos(O(m)-O(n)) + B(m,n)*sin(O(m)-O(n)));
                end
                H31(i,k) = H31(i,k) - V(m)^2*G(m,m);
            else
                H31(i,k) = V(m)* V(k+1)*(-G(m,k+1)*cos(O(m)-O(k+1)) - B(m,k+1)*sin(O(m)-O(k+1)));
            end
        end
    end
    
    % H32 - Derivative of Reactive Power Injections with V
    H32 = zeros(nqi,nbus);
    for i = 1:nqi
        m = fbus(qi(i));
        for k = 1:(nbus)
            if k == m
                for n = 1:nbus
                    H32(i,k) = H32(i,k) + V(n)*(G(m,n)*sin(O(m)-O(n)) - B(m,n)*cos(O(m)-O(n)));
                end
                H32(i,k) = H32(i,k) - V(m)*B(m,m);
            else
                H32(i,k) = V(m)*(G(m,k)*sin(O(m)-O(k)) - B(m,k)*cos(O(m)-O(k)));
            end
        end
    end
    
    % H41 - Derivative of Real Power Flows with resepect to phase angles
    H41 = zeros(npf,nbus-1);
    for i = 1:npf
        m = fbus(pf(i));
        n = tbus(pf(i));
        for k = 1:(nbus-1)
            if k+1 == m
                H41(i,k) = V(m)* V(n)*(g(m,n)*sin(O(m)-O(n)) - b(m,n)*cos(O(m)-O(n)));
            elseif k+1 == n
                H41(i,k) = -V(m)* V(n)*(g(m,n)*sin(O(m)-O(n)) - b(m,n)*cos(O(m)-O(n)));
            end
        end
    end
    
    % H42 - Derivative of Real Power Flows with V
    H42 = zeros(npf,nbus);
    for i = 1:npf
        m = fbus(pf(i));
        n = tbus(pf(i));
        for k = 1:nbus
            if k == m
                H42(i,k) = -V(n)*(-G(m,n)*cos(O(m)-O(n)) - B(m,n)*sin(O(m)-O(n))) - 2*G(m,n)*V(m);
            elseif k == n
                H42(i,k) = -V(m)*(-G(m,n)*cos(O(m)-O(n)) - B(m,n)*sin(O(m)-O(n)));
            end
        end
    end
    
    % H51 - Derivative of Reactive Power Flows with resepect to phase angles
    H51 = zeros(nqf,nbus-1);
    for i = 1:nqf
        m = fbus(qf(i));
        n = tbus(qf(i));
        for k = 1:(nbus-1)
            if k+1 == m
                H51(i,k) = -V(m)* V(n)*(-G(m,n)*cos(O(m)-O(n)) - B(m,n)*sin(O(m)-O(n)));
            elseif k+1 == n
                H51(i,k) = V(m)* V(n)*(-G(m,n)*cos(O(m)-O(n)) - B(m,n)*sin(O(m)-O(n)));
            end
        end
    end
    
    % H52 - Derivative of Reactive Power Flows with V
    H52 = zeros(nqf,nbus);
    for i = 1:nqf
        m = fbus(qf(i));
        n = tbus(qf(i));
        for k = 1:nbus
            if k == m
                H52(i,k) = -V(n)*(-G(m,n)*sin(O(m)-O(n)) + B(m,n)*cos(O(m)-O(n))) - 2*V(m)*(-B(m,n)+ bsh(m,n));
            elseif k == n
                H52(i,k) = -V(m)*(-G(m,n)*sin(O(m)-O(n)) + B(m,n)*cos(O(m)-O(n)));
            end
        end
    end
    
%     % H61 - Derivative of V_R with resepect to phase angles
%     H61 = zeros(nvir,nbus-1);
%     for i = 1:nvir
%         m = fbus(vir(i));
%         for k = 1:(nbus-1)
%             if (k+1) == m
%                 H61(i,k) = -V(m)*sin(O(m));
%             end
%         end
%     end
%     
%     % H62 - Derivative of V_R with respect to V
%     H62 = zeros(nvir,nbus);
%     for i = 1:nvir
%         m = fbus(vir(i));
%         for k = 1:nbus
%             if k == m
%                 H62(i,k) = cos(O(m));
%             end
%         end
%     end
%     
%     % H71 - Derivative of V_imag with resepect to phase angles
%     H71 = zeros(nvij,nbus-1);
%     for i = 1:nvij
%         m = fbus(vij(i));
%         for k = 1:(nbus-1)
%             if (k+1) == m
%                 H71(i,k) = V(m)*cos(O(m));
%             end
%         end
%     end
% 
%     % H72 - Derivative of V_imag with respect to V
%     H72 = zeros(nvij,nbus);
%     for i = 1:nvij
%         m = fbus(vij(i));
%         for k = 1:nbus
%             if k == m
%                 H72(i,k) = sin(O(m));
%             end
%         end
%     end
    
    % H81 - Derivative of I_real with resepect to phase angles
    H81 = zeros(nifr,nbus-1);
    for i = 1:nifr
        m = fbus(ifr(i));
        n = tbus(ifr(i));
        for k = 1:(nbus-1)
            if k+1 == m
                H81(i,k) = -V(m)*( g(m,n)*sin(O(m))+b(m,n)*cos(O(m))+bsh(m)*cos(O(m)));
            elseif k+1 == n
                H81(i,k) = V(n)*( g(m,n)*sin(O(n))+b(m,n)*cos(O(n)));
            end
        end
    end
        
    
    % H82 - Derivative of I_real with respect to V 
    H82 = zeros(nifr,nbus);
    for i = 1:nifr
        m = fbus(ifr(i));
        n = tbus(ifr(i));
        for k = 1:nbus
            if k == m
                H82(i,n) = g(m,n)*cos(O(m)) - b(m,n)*sin(O(m)) - bsh(m)*sin(O(m));
            elseif k == n
                H82(i,n) = - g(m,n)*cos(O(n))+b(m,n)*sin(O(n));
            end
        end
    end
    
    % H91 - Derivative of I_imag with resepect to phase angles
    H91 = zeros(nifj,nbus-1);
    for i = 1:nifj
        m = fbus(ifj(i));
        n = tbus(ifj(i));
        for k = 1:(nbus-1)
            if (k+1) == m
                H91(i,k) = V(m) * (-b(m,n)*sin(O(m)) - g(m,n)*cos(O(m)) + bsh(m)*sin(O(m)));
            elseif (k+1) == n
                H91(i,k) = V(n)* (b(m,n)*sin(O(n))+g(m,n)*cos(O(n)));
            end
        end
    end
    
    % H92 - Derivative of I_imag with respect to V 
    H92 = zeros(nifj,nbus);
    for i = 1:nifj
        m = fbus(ifj(i));
        n = tbus(ifj(i));
        for k = 1:nbus
            if k == m
                H92(i,k) = (b(m,n)*cos(O(m)) - g(m,n)*sin(O(m)) - bsh(m)*cos(O(m)));
            elseif k == n
                H92(i,k) = (-b(m,n)*cos(O(n))+g(m,n)*sin(O(n)));
            end
        end
    end
    
    % The resulting H matrix:
    H = [H11 H12; H21 H22; H31 H32; H41 H42; H51 H52; H81 H82; H91 H92;];
    
   %% Calculating correction dx

    Gm = H'*W*H; % Gain Matrix

    dx = Gm\(H'*W*r); % State update
    x = x + dx; % State estimate

    O(2:end) = x(1:nbus-1); %phase angle
    V = x(nbus:end); %magnitude

    iter = iter + 1;
    tol = max(abs(dx));
    
end

% Final state estimate:
Del = 180/pi*O;
result = zeros(nbus,3);
for m = 1:nbus
    result(m,1) = m;
    result(m,2) = V(m);
    result(m,3) = Del(m);
end

%% Building the final h vector
h1 = V(fbus(vi),1);
h2 = zeros(npi,1);
h3 = zeros(nqi,1);
h4 = zeros(npf,1);
h5 = zeros(nqf,1);
%h6 = V(fbus(vir)).*cos(O(fbus(vir)));
%h7 = V(fbus(vij)).*sin(O(fbus(vij)));
h8 = zeros(nifr, 1);
h9 = zeros(nifj, 1);

for i = 1:npi
    m = fbus(ppi(i));
    for k = 1:nbus
        h2(i) = h2(i) + V(m)*V(k)*(G(m,k)*cos(O(m)-O(k)) + B(m,k)*sin(O(m)-O(k)));
    end
end

for i = 1:nqi
    m = fbus(qi(i));
    for k = 1:nbus
        h3(i) = h3(i) + V(m)*V(k)*(G(m,k)*sin(O(m)-O(k)) - B(m,k)*cos(O(m)-O(k)));
    end
end

for i = 1:npf
    m = fbus(pf(i));
    n = tbus(pf(i));
    h4(i) = V(m)^2*g(m,n) - V(m)*V(n)*(g(m,n)*cos(O(m)-O(n)) + b(m,n)*sin(O(m)-O(n)));
end

for i = 1:nqf
    m = fbus(qf(i));
    n = tbus(qf(i));
    h5(i) = -V(m)^2*(b(m,n)+bsh(m,n)) - V(m)*V(n)*(g(m,n)*sin(O(m)-O(n)) - b(m,n)*cos(O(m)-O(n)));
end

for i = 1:nifr
    m = fbus(ifr(i));
    n = tbus(ifr(i));
    h8(i) = ( V(m)*cos(O(m))-V(n)*cos(O(n)) )*g(m,n)  - ( V(m)*sin(O(m)) - V(n)*sin(O(n)) )*b(m,n) - bsh(1,2)*V(m)*sin(O(m));
end

for i = 1:nifj
    m = fbus(ifj(i));
    n = tbus(ifj(i));
    h9(i) = ( ( V(m)*cos(O(m))-V(n)*cos(O(n)) )*b(m,n)  + ( V(m)*sin(O(m)) - V(n)*sin(O(n)) )*g(m,n) + bsh(m,n)*V(m)*cos(O(m)));
end


%% Additional output variables:
h = [h1; h2; h3; h4; h5; h8; h9]; %final measurement estimate
r = z - h; % residual
Ohm = Ri-H*(Gm\H');
r_N = abs(r)./sqrt(abs(diag(Ohm))); %normalised residuals
J=r'*W*r; %objective function value

end